# Description: This file is used to profile the code using cProfile

import cProfile
import pstats
from pstats import SortKey

from tests.multi.examples.option import run_model  # noqa: F401

if __name__ == "__main__":
    run_model()  # Run once for cold start imports
    cProfile.run("run_model()", "restats")
    p = pstats.Stats("restats")
    p.sort_stats(SortKey.CUMULATIVE).print_stats(30)



"""
Testing Vanilla Option for all Black Scholes models.
"""

import numpy as np
import pytest
from pytest import approx

from finmc.calc.option import opt_price_mc
from finmc.utils.assets import Discounter, Forwards
from finmc.utils.bs import opt_price
from tests.localvol.dataset import (
    data_bsmc,
    data_lvmc,
    data_lvmc_fn,
    data_lvmc_grid,
)


# datasets for all models that can price a vanilla option.
@pytest.fixture(
    params=[
        data_bsmc,
        data_lvmc,
        data_lvmc_fn,
        data_lvmc_grid,
    ]
)
def data(request):
    return request.param()


@pytest.mark.parametrize("maturity", [1, 12, 36])  # in months
@pytest.mark.parametrize("strike_x", [0.01, 0.95, 1.00, 1.05, 1.2, 10])
def test_call(data, maturity, strike_x):
    """Test the price of a vanilla call option."""

    model_cls, dataset, other = data
    spot = other["spot"]
    asset_name = dataset["BS"]["ASSET"]

    strike = strike_x * spot
    maturity_yrs = maturity / 12

    model = model_cls(dataset)
    price = opt_price_mc(
        strike,
        maturity_yrs,
        "Call",
        asset_name,
        model,
    )

    # get closed form price
    discounter = Discounter(dataset["ASSETS"][dataset["BASE"]])
    df = discounter.discount(maturity_yrs)

    asset_fwds = Forwards(dataset["ASSETS"][asset_name])
    F = asset_fwds.forward(maturity_yrs)

    sigma = dataset["BS"]["VOL"]
    if callable(sigma):
        spot = asset_fwds.forward(0)
        x = np.log(strike / spot)
        sigma = float(sigma((maturity_yrs, x)))

    expected, _ = opt_price(
        strike,
        maturity_yrs,
        "Call",
        F=F,
        df=df,
        sigma=sigma,
    )
    error = (price - expected) / spot
    assert error == approx(0.0, abs=1e-3)
    print(
        f"{maturity:3} {strike_x:6.2f}: {price:11.6f} {expected:11.6f} {error * 1e4:7.2f}"
    )


if __name__ == "__main__":
    pytest.main([__file__])



# Description: Run Heston MC model with a vanilla option contract.

from finmc.calc.option import opt_price_mc
from tests.heston.dataset import data_heston_kruse


def run_model():
    """Price a vanilla option with the Heston MC model."""

    model_cls, dataset, other = data_heston_kruse()
    spot = other["spot"]

    # Create an option timetable with 1 year maturity

    model = model_cls(dataset)
    price = opt_price_mc(
        spot,
        1,
        "Call",
        "EQ",
        model,
    )

    return price


if __name__ == "__main__":
    price = run_model()
    print(f"price = {price}")



# Description: Tests for the Heston model for vanilla options.

import numpy as np
import pytest
from pytest import approx

from finmc.utils.assets import Discounter


def test_discounts():
    """Test the Discounter class."""

    # define a discount curve using the zero rate schema
    times = np.array([0.0, 1.0, 2.0, 5.0])
    zero_rates = np.array([0.04, 0.04, 0.045, 0.05])

    test_times = [0.1, 1.0, 3.0, 5.0]
    expected_logdf = [
        0.1 * 0.04,
        1.0 * 0.04,
        2.0 * 0.045 + (0.05 * 5 - 0.045 * 2) * (3 - 2) / (5 - 2),
        5.0 * 0.05,
    ]

    # define a discount curve using the zero rate schema
    zero_data = ("ZERO_RATES", np.column_stack((times, zero_rates)))

    # define a discount curve using the log discount schema
    log_dfs = -times * zero_rates
    logdf_data = ("LOG_DISCOUNTS", np.column_stack((times, log_dfs)))

    for df_data in [zero_data, logdf_data]:
        discounter = Discounter(df_data)
        schema_name = df_data[0]
        print(f"Testing {schema_name} schema")
        for time, logdf in zip(test_times, expected_logdf):
            df = discounter.discount(time)
            expected_df = np.exp(-logdf)
            assert df == approx(expected_df, abs=1e-6)
            print(f"{df:11.6f} {expected_df:11.6f} {df - expected_df:9.6f}")


if __name__ == "__main__":
    pytest.main([__file__])



"""
Utility to calculate prices of european contracts from a MC Simulation model.
"""

import numpy as np

from finmc.models.base import MCBase


def opt_price_mc(
    strike: float,
    maturity: float,
    option_type: str,
    asset_name: str,
    model: MCBase,
) -> float:
    """Calculate the price of a Vanilla European Option using MC Simulation.

    Args:
        strike: The strike price of the option.
        maturity: The time to maturity of the option in years.
        option_type: The type of the option. Either "Call" or "Put".
        asset_name: The name of the asset.
        model: The model used to simulate the asset price.

    Returns:
        The price of the option.

    Examples:
        >>> price = opt_price_mc(K, T, "Call", "SPX", model)
    """

    model.reset()
    model.advance(maturity)
    expiration_spots = model.get_value(asset_name)
    df = model.get_df()

    if option_type == "Call":
        price = np.maximum(expiration_spots - strike, 0).mean() * df
    else:
        price = np.maximum(strike - expiration_spots, 0).mean() * df
    return price


if __name__ == "__main__":
    from finmc.models.localvol import LVMC

    dataset = {
        "MC": {
            "PATHS": 100_000,
            "TIMESTEP": 1 / 10,
            "SEED": 1,
        },
        "BASE": "USD",
        "ASSETS": {
            "USD": ("ZERO_RATES", np.array([[1.0, 0.04]])),
            "SPX": ("FORWARDS", np.array([[0.0, 2900], [1.0, 3000]])),
        },
        "LV": {
            "ASSET": "SPX",
            "VOL": 0.3,
        },
    }

    model = LVMC(dataset)
    price = opt_price_mc(
        strike=2900,
        maturity=1 / 12,
        option_type="Call",
        asset_name="SPX",
        model=model,
    )
    print(f"price = {price}")



# Description: datasets for the Heston model.

from finmc.models.heston import HestonMC
from finmc.utils.assets import flat_discount, flat_fwds


def data_heston_kruse():
    """Define the dataset from Kruse."""
    asset_name = "EQ"
    spot = 100
    rate = 0.0

    dataset = {
        "MC": {
            "PATHS": 100_000,
            "TIMESTEP": 1 / 250,
            "SEED": 1,
        },
        "BASE": "USD",
        "ASSETS": {
            "USD": flat_discount(rate, 3.0),
            asset_name: flat_fwds(spot, rate, 0.0, 3.0),
        },
        "HESTON": {
            "ASSET": asset_name,
            "INITIAL_VAR": 0.09,
            "LONG_VAR": 0.06,
            "VOL_OF_VOL": 0.65,
            "MEANREV": 4.0,
            "CORRELATION": -0.90,
        },
    }

    return (
        HestonMC,
        dataset,
        {"spot": spot, "rate": rate},
    )



import numpy as np


class UniformGridInterp:
    """Helper class to interpolate, when x-array is uniformly spaced.
    This avoids the cost of index search in arbitrary x-array."""

    def __init__(self, xmin=-3.0, xmax=3.0, dx=0.01):
        """Initialize the x-range and allocate some arrays.

        Args:
            xmin: minimum x value.
            xmax: maximum x value.
            dx: spacing between x values.
        """

        self.dx = dx
        self.xmin = xmin
        self.x_vec = np.arange(xmin, xmax + dx / 2, dx)
        self.xlen = len(self.x_vec)

        # Pre-allocate arrays
        self.slope = np.zeros(self.xlen)

    def interp(self, x_vec, y_vec, out):
        """Interpolate y_vec at x_vec and store the result in out.

        Args:
            x_vec: x values to interpolate at.
            y_vec: known y values for interpolator's stored grid (x_vec).
            out: pre-allocated destination array to store the result.
        """

        # Find the left index of the interval containing x: idx = floor((x - xmin) / dx)
        # Reusing the out array as a temp array, as it is the same shape
        np.subtract(x_vec, self.xmin, out=out)
        np.divide(out, self.dx, out=out)
        idx = out.astype(int)

        # Clip the index to [0, xlen - 1]
        np.clip(idx, 0, self.xlen - 1, out=idx)

        # get slope of y in each interval
        np.subtract(y_vec[1:], y_vec[:-1], out=self.slope[:-1])
        self.slope[-1] = self.slope[-2]
        np.divide(self.slope, self.dx, out=self.slope)

        # y = (x - left) * slope + left
        np.subtract(x_vec, self.x_vec[idx], out=out)  # x - left
        np.multiply(out, self.slope[idx], out=out)  # * slope
        np.add(out, y_vec[idx], out=out)  # + left


if __name__ == "__main__":
    interp = UniformGridInterp(xmin=-2.0, xmax=2.0, dx=1.0)
    y_vec = interp.x_vec * 10  # Some arbitrary function
    print("x=", interp.x_vec)
    print("y=", y_vec)

    new_x = np.array([-0.5, 0.5])
    new_y = new_x * 0  # pre-allocate
    interp.interp(new_x, y_vec, new_y)
    print("x=", new_x)
    print("y=", new_y)



import matplotlib.pyplot as plt
import numpy as np

from finmc.models.base import MCBase


def plot_distribution(
    model: MCBase,
    asset_name: str,
    asof: float = 1.0,
    xmax: float = 3.0,
):
    """Plot the progression of an asset in a model.

    Args:
        model: The model to use.
        asset_name: The name of the asset to plot.
        asof: The asof time (years) to plot the distribution.
        xmax: The max xrange in multiple of the mean value.

    Examples:
        >>> plot_distribution(model, "SPX", 1.0)
    """

    model.reset()
    model.advance(asof)
    spots = model.get_value(asset_name)

    # Plot distribution
    fig, ax = plt.subplots(figsize=(6, 4))
    ax.hist(spots, bins=100, density=True, alpha=0.3, color="r")

    ax.set_xlim(0, xmax * spots.mean())
    for side in ["left", "right", "top"]:
        ax.spines[side].set_visible(False)
    ax.yaxis.set_visible(False)
    ax.set_xlabel(asset_name)

    plt.show()


if __name__ == "__main__":
    from finmc.models.localvol import LVMC

    # create the dataset
    dataset = {
        "MC": {"PATHS": 100_000, "TIMESTEP": 1 / 10},
        "BASE": "USD",
        "ASSETS": {
            "USD": ("ZERO_RATES", np.array([[1.0, 0.04]])),
            "SPX": ("FORWARDS", np.array([[0.0, 5600], [1.0, 5700]])),
        },
        "LV": {
            "ASSET": "SPX",
            "VOL": 0.3,
        },
    }
    # create the model and plot the progression of short rate???
    model = LVMC(dataset)
    plot_distribution(model, "SPX", 1.0)



import matplotlib.pyplot as plt
import numpy as np

from finmc.models.base import MCBase


def plot_asset(
    model: MCBase,
    asset_name: str,
    sample_idxs=np.arange(0, 3, 1),
    q_levels=np.linspace(0.02, 0.98, 25),
    times=np.linspace(0, 1, 101),
):
    """Plot the progression of an asset in a model.

    Args:
        model: The model to use.
        asset_name: The name of the asset to plot.
        sample_idxs (np.array, optional): The indices of the sample paths to plot in the foreground. Defaults to the first three.
        q_levels (np.array, optional): The quantiles to show in the background. Defaults to 25 levels from 2% to 98%.
        times (np.array, optional): The timesteps. Defaults to 1 year in 100 steps.

    Examples:
        >>> plot_asset(model, "SPX")
    """

    num_levels = len(q_levels)
    num_steps = len(times)

    samples = np.zeros((len(sample_idxs), num_steps))
    quantiles = np.zeros((len(q_levels), num_steps))

    # enumerate over the time steps and calculate the spot price
    model.reset()
    for i, t in enumerate(times):
        model.advance(t)
        spots = model.get_value(asset_name)
        quantiles[:, i] = np.quantile(spots, q_levels)
        samples[:, i] = spots[sample_idxs]

    fig, ax = plt.subplots(figsize=(8, 3))

    for i in range(num_levels >> 1):
        ax.fill_between(
            times,
            quantiles[i, :],
            quantiles[num_levels - 1 - i, :],
            color="darkred",
            alpha=1.5 / num_levels,
            edgecolor="none",
        )
    for sample in samples:
        ax.plot(times, sample)
    ax.set_xlabel("years")
    ax.set_ylabel(asset_name)
    ax.tick_params(axis="x", direction="in", pad=-15)

    for side in ["left", "bottom", "right", "top"]:
        ax.spines[side].set_visible(False)

    plt.show()


if __name__ == "__main__":
    from finmc.models.hullwhite import HullWhiteMC

    discount_data = (
        "ZERO_RATES",
        np.array([[0.5, 0.05], [1.0, 0.04], [3, 0.04]]),
    )

    dataset = {
        "MC": {"PATHS": 100_000, "TIMESTEP": 1 / 250, "SEED": 1},
        "BASE": "USD",
        "ASSETS": {"USD": discount_data},
        "HW": {
            "ASSET": "USD",
            "MEANREV": 0.1,
            "VOL": 0.03,
        },
    }
    # create the model and plot the progression of short rate???
    model = HullWhiteMC(dataset)
    plot_asset(model, "r")



# Generic MC model

from abc import ABC, abstractmethod


class MCBase(ABC):
    """Base class for a Monte-Carlo process."""

    stats: dict = {}

    def __init__(self, dataset: dict) -> None:
        self.dataset = dataset
        self.reset()

    @abstractmethod
    def reset(self):
        """The derived class must implement this method to reset the state of the model
        to time zero."""
        ...

    def set_stat(self, key: str, val):
        self.stats[key] = val

    def get_value(self, unit):
        """Return the value of the asset at the current time,
        if this asset is handled by the model, otherwise return None.
        The return value is none, float, or an np array of floats."""
        return None

    def get_df(self):
        """Return the discount factor at the current time.
        The return value is a float, or an np array of floats."""
        ...

    @abstractmethod
    def advance(self, new_time: float):
        """The derived class must implement this method to advance the state of the model
        to a new time. The model may do so in multiple time steps."""
        ...


class MCFixedStep(MCBase):
    """A Monte-Carlo process which breaks down the 'advance' step into fixed time steps
    specified by the TIMESTEP parameter."""

    def advance(self, new_time):
        while new_time > self.cur_time + self.timestep:
            self.step(self.cur_time + self.timestep)
        if new_time > self.cur_time + 1e-10:
            self.step(new_time)

    @abstractmethod
    def step(self, new_time: float):
        """The derived class must implement this method, which advances the model by a timestep equal to or
        less than the TIMESTEP parameter."""
        ...



import matplotlib.pyplot as plt
import numpy as np
from matplotlib import cm


def plot_iv(strikes, expirations, surface):
    """Plot the implied volatility surface and forward curve.

    Args:
        strikes (np.array): The strike prices.
        expirations (np.array): The expiration times.
        surface (np.array): The implied volatility surface.

    Examples:
        >>> plot_iv(strikes, expirations, surface)
    """

    X, Y = np.meshgrid(strikes, expirations)
    fig, ax = plt.subplots(subplot_kw={"projection": "3d"}, figsize=(6, 6))
    ax.plot_surface(
        X,
        Y,
        surface,
        cmap=cm.inferno,
        alpha=0.5,
    )
    # Add wireframes for each expiration
    ax.plot_wireframe(X, Y, surface, color="brown", rstride=1, cstride=0)

    (z0, z1) = ax.get_zlim()
    if z1 - z0 < 0.05:
        ax.set_zlim(z0, z0 + 0.05)

    ax.set_xlabel("Strike (K)")
    ax.set_ylabel("Maturity (years)")
    plt.show()


def plot_atmvols(expirations, atm_vols):
    """Plot the implied volatility surface and forward curve.

    Args:
        expirations (np.array): The expiration times.
        atm_vols (np.array): The ATM volatilities.

    Examples:
        >>> plot_atmvols(expirations, atm_vols)
    """

    # Add atm vol curve
    fig, ax = plt.subplots(figsize=(5, 2))
    ax.plot(
        expirations, atm_vols, color="brown", label="Forward Curve", marker="o"
    )

    (y0, y1) = ax.get_ylim()
    if y1 - y0 < 0.05:
        ax.set_ylim(y0 - 0.01, y0 + 0.05)

    ax.set_xlabel("Maturity (years)")
    ax.set_ylabel("ATM Vol")
    plt.show()



# Description: Tests for a custom MC Hull White model with ZCB.

import numpy as np

from finmc.models.hullwhite import HullWhiteMC


def data_hwmc():
    """Data for the MC HW model."""

    # define dataset
    # first define the discount curve
    times = np.array([0.0, 1.0, 2.0, 5.0, 10.0])
    term_rates = np.array([0.04, 0.04, 0.045, 0.05, 0.05])
    discount_data = ("ZERO_RATES", np.column_stack((times, term_rates)))

    dataset = {
        "MC": {
            "PATHS": 100_000,
            "TIMESTEP": 1 / 250,
            "SEED": 1,
        },
        "BASE": "USD",
        "ASSETS": {"USD": discount_data},
        "HW": {
            "ASSET": "USD",
            "MEANREV": 0.1,
            "VOL": 0.03,
        },
    }

    return HullWhiteMC, dataset, {}



import numpy as np
from finmc.models.base import MCFixedStep
from finmc.utils.assets import Discounter, Forwards
from numpy.random import SFC64, Generator


class BSMC(MCFixedStep):
    """Class for the multi asset Black Scholes model."""

    def reset(self):
        """Fetch any information from the dataset or timetable, that needs to be stored into self,
        to facilitate the 'advance' method to run as quickly as possible."""
        # common MC parameters
        self.n = self.dataset["MC"]["PATHS"]
        self.timestep = self.dataset["MC"]["TIMESTEP"]

        # fetch the model parameters from the dataset
        modeldata = self.dataset["BSM"]
        self.cov = modeldata["COV"]

        # Create a dictionary of assets, so that we can easily access the idx and spot from string name.
        # Also create ordered lists of forwards and variances for the assets.
        self.assets = {}
        self.fwds = []
        self.vars = []
        for idx, asset in enumerate(modeldata["ASSETS"]):
            fwd = Forwards(self.dataset["ASSETS"][asset])
            self.assets[asset] = {"idx": idx, "spot": fwd.forward(0)}
            self.fwds.append(fwd)
            self.vars.append(self.cov[idx][idx])

        self.discounter = Discounter(
            self.dataset["ASSETS"][self.dataset["BASE"]]
        )

        # Initialize the RNG and the x_vec (log process for each asset)

        self.rng = Generator(SFC64(self.dataset["MC"]["SEED"]))
        self.x_vec = np.zeros((len(self.assets), self.n))

        self.cur_time = 0

    def step(self, new_time):
        """Update x_vec in place when we move simulation by time dt."""

        dt = new_time - self.cur_time

        drifts = np.array(
            [fwd.rate(new_time, self.cur_time) for fwd in self.fwds]
        )
        drifts -= np.array(self.vars) / 2.0

        # generate the random numbers and advance the log stock process
        dz = self.rng.multivariate_normal(
            drifts * dt, self.cov * dt, self.n
        ).transpose()

        self.x_vec += dz

        self.cur_time = new_time

    def get_value(self, unit):
        """If this asset is part of the model, return its value using the simulated array.
        For any other asset that may exist in the timetable, return the default implementation in qablet base model.
        """

        info = self.assets.get(unit)
        if info:
            return info["spot"] * np.exp(self.x_vec[info["idx"]])

    def get_df(self):
        return self.discounter.discount(self.cur_time)



# Description: Run Heston MC model with a vanilla option contract.

from tests.multi.dataset import data_three_asset
from tests.multi.tests.test_basket import _basket_price_mc


def run_model():
    """Price a vanilla option with the Heston MC model."""

    model_cls, dataset, other = data_three_asset()
    model = model_cls(dataset)
    spots = other["spots"]
    price = _basket_price_mc(model, maturity=1.0, strike=1.0, spots=spots)

    return price


if __name__ == "__main__":
    price = run_model()
    print(f"price = {price}")



# Description: Tests for the multi asset black scholes model for vanilla options.

import pytest
from pytest import approx

from finmc.utils.assets import Discounter, Forwards
from tests.multi.dataset import data_three_asset
import numpy as np
from finmc.utils.bs import opt_price


def _basket_price_mc(model, maturity, strike, spots):
    """Calculate the price of a equal weight basket call option using simulation."""
    model.reset()
    model.advance(maturity)
    pay = -strike
    for asset_name, spot in spots.items():
        pay += 0.5 * model.get_value(asset_name) / spot

    df = model.get_df()
    price = np.maximum(pay, 0.0).mean() * df
    return price


def _basket_price_bs(dataset, maturity, strike, spots):
    """Calculate the price of a equal weight basket call option using closed form."""
    ccy = dataset["BASE"]
    discounter = Discounter(dataset["ASSETS"][ccy])
    df = discounter.discount(maturity)

    F = 0.0
    for asset_name, spot in spots.items():
        asset_fwds = Forwards(dataset["ASSETS"][asset_name])
        Fi = asset_fwds.forward(maturity)
        F += 0.5 * Fi / spot

    sigma = np.sqrt(dataset["BSM"]["COV"].sum()) / 2.0
    expected, _ = opt_price(
        strike,
        maturity,
        "Call",
        F=F,
        df=df,
        sigma=sigma,
    )
    return expected


@pytest.fixture(scope="module")
def data():
    return data_three_asset()


@pytest.mark.parametrize("maturity", [0.1, 2.0])
@pytest.mark.parametrize("strike", [0.01, 0.95, 1.00, 1.05, 1.2])
def test_call(data, maturity, strike):
    """Test the price of a basket call option."""

    model_cls, dataset, other = data
    spots = other["spots"]

    model = model_cls(dataset)

    # Calculate (two asset) basket price using MC
    price = _basket_price_mc(model, maturity, strike, spots)

    # Calculate (two asset) basket price using BS
    expected = _basket_price_bs(dataset, maturity, strike, spots)

    error = price - expected
    contract = f"Call {maturity:4.2f} {strike:6.2f}"
    assert error == approx(0.0, abs=1e-3)
    print(f"{contract}: {price:11.6f} {expected:11.6f} {error * 1e4:7.2f}")


if __name__ == "__main__":
    pytest.main([__file__])



# Helper functions for Heston model, which includes calibration,
# and closed form pricing for special cases.

# Based on https://github.com/cantaro86/Financial-Models-Numerical-Methods/

from functools import partial

import numpy as np
from scipy.integrate import quad


def cf_heston(u, t, v0, mu, kappa, theta, sigma, rho):
    """
    Heston characteristic function as proposed by Schoutens (2004)
    """
    xi = kappa - sigma * rho * u * 1j
    d = np.sqrt(xi**2 + sigma**2 * (u**2 + 1j * u))
    g1 = (xi + d) / (xi - d)
    g2 = 1 / g1
    cf = np.exp(
        1j * u * mu * t
        + (kappa * theta)
        / (sigma**2)
        * ((xi - d) * t - 2 * np.log((1 - g2 * np.exp(-d * t)) / (1 - g2)))
        + (v0 / sigma**2)
        * (xi - d)
        * (1 - np.exp(-d * t))
        / (1 - g2 * np.exp(-d * t))
    )
    return cf


def Q1(k, cf, right_lim):
    """
    P(X<k) - Probability to be in the money under the stock numeraire.
    cf: characteristic function
    right_lim: right limit of integration
    """

    def integrand(u):
        return np.real(
            (np.exp(-u * k * 1j) / (u * 1j))
            * cf(u - 1j)
            / cf(-1.0000000000001j)
        )

    return 1 / 2 + 1 / np.pi * quad(integrand, 1e-15, right_lim, limit=2000)[0]


def Q2(k, cf, right_lim):
    """
    P(X<k) - Probability to be in the money under the money market numeraire
    cf: characteristic function
    right_lim: right limit of integration
    """

    def integrand(u):
        return np.real(np.exp(-u * k * 1j) / (u * 1j) * cf(u))

    return 1 / 2 + 1 / np.pi * quad(integrand, 1e-15, right_lim, limit=2000)[0]


def opt_price_heston(
    K,  # strike
    T,  # option maturity in years
    S0,  # initial spot
    r,  # risk free rate
    mu,  # stock drift
    heston_params,
):
    """Calculate the price of a Vanilla European Option using Heston Model characteristic function."""
    # TODO feat: handle time dependent params.

    cf_reduced = partial(
        cf_heston,
        t=T,
        v0=heston_params["INITIAL_VAR"],
        mu=mu,
        theta=heston_params["LONG_VAR"],
        sigma=heston_params["VOL_OF_VOL"],
        kappa=heston_params["MEANREV"],
        rho=heston_params["CORRELATION"],
    )
    limit_max = 1000
    k = np.log(K / S0)  # log strike
    price = S0 * np.exp((mu - r) * T) * Q1(
        k, cf_reduced, limit_max
    ) - K * np.exp(-r * T) * Q2(k, cf_reduced, limit_max)
    return price, {}



# Description: datasets for the LV MC and BS FD model.

import numpy as np

from finmc.models.localvol import BSMC, LVMC


def _create_interp_two_step(times, strikes, vols):
    t_len = len(times)

    def interp(args):
        t, x_vec = args
        i_left = np.searchsorted(times, t, side="left")
        i_right = np.searchsorted(times, t, side="right")

        if i_left == i_right:
            if i_right == 0:
                y = vols[0]
            elif i_right == t_len:
                y = vols[-1]
            else:
                i_left -= 1
                t_right = times[i_right]
                t_left = times[i_left]
                den = t_right - t_left
                y = (
                    vols[i_left]
                    + (vols[i_right] - vols[i_left]) * (t - t_left) / den
                )
        else:
            y = vols[i_left]
        return np.interp(x_vec, strikes, y)

    return interp


# Description: datasets for the LVMC model.
def _bs_vol(points):
    """A simple BS implied vol function with no skew."""
    (t, x_vec) = points
    var = 0.04 + 0.04 * (1 - np.exp(-0.5 * t))  # No skew
    return np.sqrt(var)


def _local_vol(points):
    """A local vol function derived from the above BS Vol function."""
    (t, x_vec) = points
    f = np.exp(-0.5 * t)
    var = 0.04 + 0.04 * (1 - f + 0.5 * t * f)
    return np.sqrt(var)


def _interp(volfn):
    """Return a RegularGridInterpolator for the vol function."""
    num_t = 10
    num_x = 20
    times = np.linspace(0.0, 3.0, num_t)
    strikes = np.linspace(-5.0, 5.0, num_x)
    vi = np.zeros((num_t, num_x))
    for i in range(num_t):
        vi[i, :] = volfn((times[i], strikes))
    return _create_interp_two_step(times, strikes, vi)


def data_lvmc():
    """Data for the LVMC model."""

    # define dataset
    # first define the discount curve and forward curve
    times = np.array([0.0, 1.0, 2.0, 5.0, 10.0])
    term_rates = np.array([0.04, 0.04, 0.045, 0.05, 0.05])
    div_rate = 0.01
    discount_data = ("ZERO_RATES", np.column_stack((times, term_rates)))

    asset_name = "SPX"
    spot = 2900
    fwds = spot * np.exp((term_rates - div_rate) * times)
    fwd_data = ("FORWARDS", np.column_stack((times, fwds)))

    dataset = {
        "MC": {
            "PATHS": 100_000,
            "TIMESTEP": 1 / 10,
            "SEED": 1,
        },
        "BASE": "USD",
        "ASSETS": {"USD": discount_data, asset_name: fwd_data},
        "BS": {
            "ASSET": "SPX",
            "VOL": 0.3,
        },  # For closed form evaluation
        "LV": {
            "ASSET": "SPX",
            "VOL": 0.3,
        },
    }

    return LVMC, dataset, {"spot": spot}


def data_lvmc_grid():
    """Test when vol is a RegularGridInterpolator."""
    model_cls, dataset, other = data_lvmc()
    dataset["BS"]["VOL"] = _interp(_bs_vol)
    dataset["LV"]["VOL"] = _interp(_local_vol)
    dataset["MC"]["TIMESTEP"] = 1 / 250

    return model_cls, dataset, other


def data_lvmc_fn():
    """Test when vol is a function."""
    model_cls, dataset, other = data_lvmc()
    dataset["BS"]["VOL"] = _bs_vol
    dataset["LV"]["VOL"] = _local_vol
    dataset["MC"]["TIMESTEP"] = 1 / 250

    return model_cls, dataset, other


def data_bsmc():
    """Test when vol is a RegularGridInterpolator."""
    _, dataset, other = data_lvmc()

    return BSMC, dataset, other



# Custom Monte Carlo Pricer for a HullWhite Model

from math import sqrt

import numpy as np
from numpy.random import SFC64, Generator

from finmc.models.base import MCFixedStep
from finmc.utils.assets import Discounter
from finmc.utils.mc import antithetic_normal


# Define a class for the state of a single asset HullWhite MC process
class HullWhiteMC(MCFixedStep):
    def reset(self):
        self.n = self.dataset["MC"]["PATHS"]
        self.timestep = self.dataset["MC"]["TIMESTEP"]

        # create a random number generator
        self.rng = Generator(SFC64(self.dataset["MC"].get("SEED")))

        self.asset = self.dataset["HW"]["ASSET"]
        self.asset_fwd = Discounter(self.dataset["ASSETS"][self.asset])

        self.meanrev = self.dataset["HW"]["MEANREV"]
        self.vol = self.dataset["HW"]["VOL"]

        # We will reduce time spent in memory allocation by creating a tmp arrays
        self.tmp_vec = np.empty(self.n, dtype=np.float64)

        # Initialize the arrays for the processes
        self.x_vec = np.zeros(self.n, dtype=np.float64)  # x
        self.r_vec = np.zeros(self.n, dtype=np.float64)  # r (short rate)
        self.df_vec = np.ones(self.n, dtype=np.float64)  # discount factors
        fwd_rate = self.asset_fwd.rate(self.timestep, 0)
        np.add(self.x_vec, fwd_rate, out=self.r_vec)

        self.cur_time = 0

    def step(self, new_time):
        """Update x_vec, v_vec in place when we move simulation by time dt."""
        dt = new_time - self.cur_time

        sqrtdt = sqrt(dt)

        # update the current value of x
        # first term: x -= a * x * dt
        np.multiply(self.x_vec, self.meanrev * dt, out=self.tmp_vec)
        np.subtract(self.x_vec, self.tmp_vec, out=self.x_vec)

        # second term: x += vol * dW
        antithetic_normal(self.rng, self.n, sqrtdt * self.vol, self.tmp_vec)
        np.add(self.x_vec, self.tmp_vec, out=self.x_vec)

        # r = x + alpha
        fwd_rate = self.asset_fwd.rate(new_time, self.cur_time)
        temp = (self.vol / self.meanrev) * (
            1 - np.exp(-self.meanrev * new_time)
        )
        alpha = fwd_rate + temp * temp / 2
        np.add(self.x_vec, alpha, out=self.r_vec)

        self.cur_time = new_time

        np.multiply(self.df_vec, np.exp(-self.r_vec * dt), out=self.df_vec)

    def get_df(self):
        """Return the discount factor at the current time."""
        return self.df_vec  # discount factor

    def get_value(self, unit):
        """Return the value of the unit at the current time."""
        if unit == "r":  # ad-hoc for now, convention to be reviewed
            return self.r_vec



# Monte Carlo Pricer for Heston Model

from math import sqrt

import numpy as np
from numpy.random import SFC64, Generator

from finmc.models.base import MCFixedStep
from finmc.utils.assets import Discounter, Forwards
from finmc.utils.mc import antithetic_normal


# Define a class for the state of a single asset Heston MC process
class HestonMC(MCFixedStep):
    def reset(self):
        self.n = self.dataset["MC"]["PATHS"]
        self.timestep = self.dataset["MC"]["TIMESTEP"]

        # create a random number generator
        self.rng = Generator(SFC64(self.dataset["MC"].get("SEED")))

        self.asset = self.dataset["HESTON"]["ASSET"]
        self.asset_fwd = Forwards(self.dataset["ASSETS"][self.asset])
        self.discounter = Discounter(
            self.dataset["ASSETS"][self.dataset["BASE"]]
        )

        self.heston_params = (
            self.dataset["HESTON"]["LONG_VAR"],
            self.dataset["HESTON"]["VOL_OF_VOL"],
            self.dataset["HESTON"]["MEANREV"],
            self.dataset["HESTON"]["CORRELATION"],
        )
        self.v0 = self.dataset["HESTON"]["INITIAL_VAR"]

        # We will reduce time spent in memory allocation by creating arrays in advance
        # and reusing them in the `advance` function which is called repeatedly.
        # though the values from one timestep are not reused in the next.
        self.tmp_vec = np.empty(self.n, dtype=np.float64)
        self.dz1_vec = np.empty(self.n, dtype=np.float64)
        self.dz2_vec = np.empty(self.n, dtype=np.float64)
        self.vol_vec = np.empty(self.n, dtype=np.float64)
        self.vplus_vec = np.empty(self.n, dtype=np.float64)

        # Initialize the arrays
        self.x_vec = np.zeros(self.n)  # processes x (log stock)
        self.v_vec = np.full(self.n, self.v0)  # processes v (variance)
        self.cur_time = 0

    def step(self, new_time):
        """Update x_vec, v_vec in place when we move simulation by time dt."""
        dt = new_time - self.cur_time

        (theta, vvol, meanrev, corr) = self.heston_params
        sqrtdt = sqrt(dt)

        # To improve preformance we will break up the operations into np.multiply,
        # np.add, etc. and use the `out` parameter to avoid creating temporary arrays.

        # calculate dz1 = normal(0,1) * sqrtdt
        antithetic_normal(self.rng, self.n, sqrtdt, self.dz1_vec)

        # calculate dz2 = normal(0,1) * sqrtdt * sqrt(1 - corr * corr) + corr * dz1
        corr_comp = sqrtdt * sqrt(1 - corr * corr)
        antithetic_normal(self.rng, self.n, corr_comp, self.dz2_vec)
        np.multiply(corr, self.dz1_vec, out=self.tmp_vec)  # second term
        np.add(self.dz2_vec, self.tmp_vec, out=self.dz2_vec)

        # vol = sqrt(max(v, 0))
        np.maximum(0.0, self.v_vec, out=self.vplus_vec)
        np.sqrt(self.vplus_vec, out=self.vol_vec)

        # update the current value of x (log Stock process)
        # first term: x -= vol * vol * dt/2
        np.divide(self.vplus_vec, 2, out=self.tmp_vec)
        np.multiply(self.tmp_vec, dt, out=self.tmp_vec)
        np.subtract(self.x_vec, self.tmp_vec, out=self.x_vec)

        # second term: x += vol * dz1
        np.multiply(self.vol_vec, self.dz1_vec, out=self.tmp_vec)
        np.add(self.x_vec, self.tmp_vec, out=self.x_vec)

        # update the current value of v (variance process)
        # first term: v += meanrev * (theta - v) * dt
        np.subtract(theta, self.v_vec, out=self.tmp_vec)
        np.multiply(self.tmp_vec, (meanrev * dt), out=self.tmp_vec)
        np.add(self.v_vec, self.tmp_vec, out=self.v_vec)

        # second term: v += vvol * vol * dz2
        np.multiply(vvol, self.vol_vec, out=self.tmp_vec)
        np.multiply(self.tmp_vec, self.dz2_vec, out=self.tmp_vec)
        np.add(self.v_vec, self.tmp_vec, out=self.v_vec)

        self.cur_time = new_time

    def get_value(self, unit):
        """Return the value of the unit at the current time."""
        if unit == self.asset:
            return self.asset_fwd.forward(self.cur_time) * np.exp(self.x_vec)

    def get_df(self):
        return self.discounter.discount(self.cur_time)



# Description: Tests for the Heston model for vanilla options.

import pytest
from pytest import approx

from finmc.calc.option import opt_price_mc
from finmc.utils.assets import Discounter, Forwards
from tests.heston.dataset import data_heston_kruse
from tests.heston.helpers import opt_price_heston


@pytest.fixture(scope="module")
def data():
    return data_heston_kruse()


@pytest.mark.parametrize("maturity", [0.1, 1.0, 3.0])
@pytest.mark.parametrize("strike_x", [0.01, 0.95, 1.00, 1.05, 1.2])
def test_call(data, maturity, strike_x):
    """Test the price of a vanilla call option."""

    model_cls, dataset, other = data
    spot = other["spot"]
    asset_name = dataset["HESTON"]["ASSET"]
    ccy = dataset["BASE"]
    strike = strike_x * spot

    model = model_cls(dataset)

    price = opt_price_mc(
        strike,
        maturity,
        "Call",
        asset_name,
        model,
    )

    # get closed form price
    discounter = Discounter(dataset["ASSETS"][ccy])
    asset_fwds = Forwards(dataset["ASSETS"][asset_name])
    r = discounter.rate(maturity)
    S0 = asset_fwds.forward(0)
    mu = asset_fwds.rate(maturity)

    expected_price, _ = opt_price_heston(
        strike,
        maturity,
        S0=S0,
        r=r,
        mu=mu,
        heston_params=dataset["HESTON"],
    )
    error = (price - expected_price) / spot
    contract = f"Call {maturity:4.2f} {strike:6.0f}"
    assert error == approx(0.0, abs=1e-3)
    print(
        f"{contract}: {price:11.6f} {expected_price:11.6f} {error * 1e4:7.2f}"
    )


if __name__ == "__main__":
    pytest.main([__file__])



"""
Utility to calculate implied volatility using closed form Black-Scholes model.
"""

import numpy as np
from scipy.stats import norm

N = norm.cdf
N_prime = norm.pdf


def d1_d2(
    K,  # strike
    T,  # option maturity in years
    F,
    vol,
):
    """Calculate d1, d2 from Black Scholes."""
    d1 = (np.log(F / K) + (vol * vol / 2) * T) / (vol * np.sqrt(T))
    d2 = d1 - vol * np.sqrt(T)
    return d1, d2


def impliedvol(target, F, K, T, is_call):
    """Find the implied volatility of a European option using the Black-Scholes model.

    Args:
        target (float): The target value of the option, undiscounted.
        F (float): The forward price of the underlying.
        K (float): The strike price of the option.
        T (float): The time to maturity of the option in years.
        is_call (bool): True if the option is a call, False if put.

    """
    if target < 1e-4 * F:
        return None
    MAX_ITERATIONS = 200
    PRECISION = 1.0e-5
    vol = 0.5
    for i in range(0, MAX_ITERATIONS):
        d1, d2 = d1_d2(K, T, F, vol)

        if is_call:
            price = F * N(d1) - K * N(d2)
        else:
            price = K * N(-d2) - F * N(-d1)
        vega = F * N_prime(d1) * np.sqrt(T)
        diff = target - price  # our root
        if abs(diff) < PRECISION:
            return vol
        vol = vol + diff / vega  # f(x) / f'(x)

    return vol


def opt_price(
    K,  # strike
    T,  # option maturity in years
    option_type,  # "Call" or "Put"
    F,  # forward
    df,  # discount factor
    sigma,  # volatility
):
    """Calculate the price of a Vanilla European Option using Closed from Black Scholes."""

    d1, d2 = d1_d2(K, T, F, sigma)

    if option_type == "Call":
        price = F * N(d1) - K * N(d2)
    else:
        price = K * N(-d2) - F * N(-d1)

    return price * df, {}


def digital_price(
    K,  # strike
    T,  # option maturity in years
    option_type,  # "Call" or "Put"
    F,  # forward
    df,  # discount factor
    sigma,  # volatility
    apply_discounter=True,
):
    """Calculate the price of a Digital European Call Option using Closed from Black Scholes."""

    _, d2 = d1_d2(K, T, F, sigma)

    if option_type == "Call":
        price = N(d2)
    else:
        price = N(-d2)

    if apply_discounter:
        price = price * df
    return price, {}


def dupire_local_var(dt: float, dk: float, k, w_vec_prev, w_vec):
    """Calculate local variance (vol**2) from t0 to t1, given an array of strikes k, uniformly spaced by dk,
    and the total variances (w = T * vol**2) at time t0 and t1.
    The vectors k, w_vec_prev, and w_vec must be of the same length.
    The returned vector is two elements shorter.

    Args:
        dt: time step, i.e. t1 - t0.
        dk: strike step
        k: np array of log strikes, uniformly spaced by dk
        w_vec_prev: np array of total variance at time t0
        w_vec: np array total variance at time t1

    """

    w = (w_vec + w_vec_prev) / 2
    # time derivative of w
    wt = (w_vec - w_vec_prev) / dt

    # strike derivatives of w
    wk = (w[2:] - w[:-2]) / (2 * dk)
    wkk = (w[2:] + w[:-2] - 2 * w[1:-1]) / (dk**2)

    # drop the extra points at top and bottom
    w_ = w[1:-1]
    wt_ = wt[1:-1]

    # apply Dupire's formula
    return wt_ / (
        1
        - k / w_ * wk
        + 1 / 4 * (-1 / 4 - 1 / w_ + k**2 / w_**2) * (wk) ** 2
        + 1 / 2 * wkk
    )


if __name__ == "__main__":
    iv = impliedvol(8.58, 5508.3, 6400, 0.0833, True)



# Description: datasets for the Heston model.

from finmc.models.multi import BSMC
from finmc.utils.assets import flat_discount, flat_fwds
import numpy as np


def data_three_asset():
    """Define a three asset dataset."""
    spots = {
        "NVDA": 116.00,
        "INTC": 21.84,
    }
    rate = 0.03

    # Asset forwards and discount curve
    assets = {k: flat_fwds(v, rate, 0.0, 3.0) for k, v in spots.items()}
    assets["USD"] = flat_discount(rate, 3.0)

    # Covariance matrix
    cov = np.array(
        [
            [0.09, 0.03],
            [0.03, 0.04],
        ]
    )

    # Complete dataset
    dataset = {
        "MC": {
            "PATHS": 100_000,
            "TIMESTEP": 1 / 10,
            "SEED": 1,
        },
        "BASE": "USD",
        "ASSETS": assets,
        "BSM": {
            "ASSETS": list(spots.keys()),
            "COV": cov,
        },
    }

    return (
        BSMC,
        dataset,
        {"spots": spots, "rate": rate},
    )



# Description: Run MSFD and LVMC models with a vanilla option contract.
from finmc.calc.option import opt_price_mc
from tests.localvol.dataset import data_lvmc_fn, data_lvmc_grid


def run_model(type="grid"):
    """Price a vanilla option with the LV MC model."""

    if type == "grid":
        model_cls, dataset, other = data_lvmc_grid()
    elif type == "fn":
        model_cls, dataset, other = data_lvmc_fn()

    spot = other["spot"]
    strike = spot
    maturity = 1

    asset_name = dataset["LV"]["ASSET"]

    # get price from timetable
    model = model_cls(dataset)
    price = opt_price_mc(
        strike,
        maturity,
        "Call",
        asset_name,
        model,
    )
    return price


if __name__ == "__main__":
    price = run_model()
    print(f"price = {price}")



"""
Utility to calculate option prices or implied volatility surface from a MC Simulation, for any model.
This is useful for calibrating the model to Black-Scholes option prices.
"""

import numpy as np

from finmc.models.base import MCBase
from finmc.utils.bs import impliedvol


def _iv_strike(spots, strikes, fwd: float, exp: float):
    # Use a call option for strikes above forward, a put option otherwise
    is_call = strikes > fwd
    is_call_c = is_call[..., None]  # Turn into a column vector

    # calculate prices (value as of expiration date)
    strikes_c = strikes[..., None]  # Turn into a column vector
    pay = np.where(
        is_call_c,
        spots - strikes_c,
        strikes_c - spots,
    )
    prices = np.maximum(pay, 0).mean(axis=1)

    # calculate implied vols
    return [
        impliedvol(p, fwd, k, exp, ic)
        for p, k, ic in zip(prices, strikes, is_call)
    ]


def iv_surface_mc(
    strikes,
    expirations,  # in years, increasing order
    asset_name: str,
    model: MCBase,
    is_log: bool = False,
):
    """Calculate the implied volatility surface using MC Simulation.

    Args:
        strikes: The strikes of the options.
        expirations: The expirations of the options in years.
        asset_name: The name of the asset.
        model: The model used to simulate the asset price.
        is_log: if True, the strikes are in log, i.e. log(K/F).

    Returns:
        The implied volatility surface as a 2D numpy array with a row for each strike and a column for each expiration.

    Examples:
        >>> strikes = np.linspace(2900, 3100, 3)
        >>> expirations = [0.25, 0.5, 1]
        >>> surface = iv_surface_mc(strikes, expirations, "SPX", model)
    """

    iv_mat = np.zeros((len(expirations), len(strikes)))

    model.reset()
    for i, exp in enumerate(expirations):
        model.advance(exp)
        spots = model.get_value(asset_name)
        fwd = spots.mean()

        # calculate implied vols
        if is_log:
            iv_mat[i, :] = _iv_strike(spots, fwd * np.exp(strikes), fwd, exp)
        else:
            iv_mat[i, :] = _iv_strike(spots, strikes, fwd, exp)
    return iv_mat


def atmvols_mc(
    expirations,  # in years, increasing order
    asset_name: str,
    model: MCBase,
):
    """Calculate the atmvols using MC Simulation.

    Args:
        expirations: The expirations of the options in years.
        asset_name: The name of the asset.
        model: The model used to simulate the asset price.

    Returns:
        A tuple containing the ATM volatilities, and the forward prices.

    Examples:
        >>> strikes = np.linspace(2900, 3100, 3)
        >>> expirations = [0.25, 0.5, 1]
        >>> atm_vols, fwds = atmvols_mc(expirations, "SPX", model)
    """

    iv_atm = []
    fwds = []
    model.reset()
    for i, exp in enumerate(expirations):
        model.advance(exp)
        spots = model.get_value(asset_name)
        fwd = spots.mean()

        # calculate atm vols
        atm_call = np.maximum(spots - fwd, 0).mean()
        # calculate implied vols and fwds
        fwds.append(fwd)
        iv_atm.append(impliedvol(atm_call, fwd, fwd, exp, True))
    return np.array(iv_atm), np.array(fwds)


if __name__ == "__main__":
    import pandas as pd

    from finmc.models.localvol import LVMC

    # create the dataset
    dataset = {
        "MC": {"PATHS": 100_000, "TIMESTEP": 1 / 10},
        "BASE": "USD",
        "ASSETS": {
            "USD": ("ZERO_RATES", np.array([[1.0, 0.04]])),
            "SPX": ("FORWARDS", np.array([[0.0, 2900], [1.0, 3000]])),
        },
        "LV": {
            "ASSET": "SPX",
            "VOL": 0.3,
        },
    }

    # create the model and calculate the implied volatility surface
    model = LVMC(dataset)
    strikes = np.linspace(2900, 3100, 3)
    expirations = [1 / 12, 1 / 6, 1 / 4, 1 / 2, 1]
    surface = iv_surface_mc(
        strikes,
        expirations,
        asset_name="SPX",
        model=model,
    )
    # print the surface as a DataFrame
    df = pd.DataFrame(surface, columns=strikes, index=expirations)
    print(f"surface:\n{df}")

    atm_vols, fwds = atmvols_mc(
        expirations,
        asset_name="SPX",
        model=model,
    )

    print(f"atm_vols:\n{atm_vols}")
    print(f"fwds:\n{fwds}")



# Utility classes and functions for models.

import numpy as np
from scipy import interpolate


# Define a class for discount factors and rates.
class Discounter:
    """A class for discount factors and rates."""

    def __init__(self, discount_data):
        data_type, data = discount_data
        if data[0, 0] > 0.0:
            data = np.insert(data, [0], [0.0, 0.0], axis=0)
        if data_type == "LOG_DISCOUNTS":
            # the columns are times and log discounts
            times = data[:, 0]
            log_discounts = data[:, 1]
        elif data_type == "ZERO_RATES":
            # the columns are times and zero rates
            times = data[:, 0]
            zero_rates = data[:, 1]
            log_discounts = -zero_rates * times

        self.log_discount_fn = interpolate.interp1d(times, log_discounts)

    def rate(self, end, start=0):
        ld_end, ld_start = self.log_discount_fn([end, start])

        return (ld_start - ld_end) / (end - start)

    def discount(self, t):
        return np.exp(self.log_discount_fn(t))


# Define a class for forwards of an asset.
class Forwards:
    """A class for forwards and forward rates."""

    def __init__(self, forwards_data):
        _, data = forwards_data

        times = data[:, 0]
        fwds = data[:, 1]

        self.log_forward_fn = interpolate.interp1d(times, np.log(fwds))

    def rate(self, end, start=0):
        try:
            ld_end, ld_start = self.log_forward_fn([end, start])
        except ValueError as e:
            raise ValueError(
                f"Error calculating forward rate from time {start} to {end}, caused by \n {e}. \n Perhaps the time range in forwards/discount data is too short."
            )
        return (ld_end - ld_start) / (end - start)

    def forward(self, t):
        return np.exp(self.log_forward_fn(t))


def flat_discount(rate, max_time):
    """Create a flat discount curve."""
    return ("ZERO_RATES", np.array([[max_time, rate]]))


def flat_fwds(spot, rate, div_rate, max_time):
    """Create a flat forward curve."""
    times = np.array([0.0, max_time])
    fwds = spot * np.exp((rate - div_rate) * times)
    return ("FORWARDS", np.column_stack((times, fwds)))



from math import sqrt

import numpy as np
from numpy.random import SFC64, Generator

from finmc.models.base import MCFixedStep
from finmc.utils.assets import Discounter, Forwards
from finmc.utils.mc import antithetic_normal


# Define a single asset Black Scholes process with a flat volatility
class BSMC(MCFixedStep):
    def reset(self):
        # fetch the model parameters from the dataset
        self.n = self.dataset["MC"]["PATHS"]
        self.timestep = self.dataset["MC"]["TIMESTEP"]

        self.asset = self.dataset["LV"]["ASSET"]
        self.asset_fwd = Forwards(self.dataset["ASSETS"][self.asset])
        self.vol = self.dataset["BS"]["VOL"]
        self.discounter = Discounter(
            self.dataset["ASSETS"][self.dataset["BASE"]]
        )

        # Initialize rng and any arrays
        self.rng = Generator(SFC64(self.dataset["MC"].get("SEED")))
        self.dz_vec = np.empty(self.n, dtype=np.float64)

        self.x_vec = np.zeros(self.n)  # process x (log stock)
        self.cur_time = 0

    def step(self, new_time):
        """Update x_vec in place when we move simulation by time dt."""

        dt = new_time - self.cur_time

        antithetic_normal(self.rng, self.n, self.vol * sqrt(dt), self.dz_vec)

        self.x_vec += (-self.vol * self.vol / 2.0) * dt + self.dz_vec

        self.cur_time = new_time

    def get_value(self, unit):
        """Return the value of the modeled asset at the current time."""
        if unit == self.asset:
            return self.asset_fwd.forward(self.cur_time) * np.exp(self.x_vec)

    def get_df(self):
        return self.discounter.discount(self.cur_time)


# Define a class for the state of a single asset BS Local Vol MC process
class LVMC(MCFixedStep):
    def reset(self):
        # fetch the model parameters from the dataset
        self.n = self.dataset["MC"]["PATHS"]
        self.timestep = self.dataset["MC"]["TIMESTEP"]

        self.asset = self.dataset["LV"]["ASSET"]
        self.asset_fwd = Forwards(self.dataset["ASSETS"][self.asset])
        self.vol = self.dataset["LV"]["VOL"]
        self.discounter = Discounter(
            self.dataset["ASSETS"][self.dataset["BASE"]]
        )

        # Create rng and tmp arrays
        self.rng = Generator(SFC64(self.dataset["MC"].get("SEED")))
        self.dz_vec = np.empty(self.n, dtype=np.float64)
        self.tmp = np.empty(self.n, dtype=np.float64)

        # Initialize the process
        self.x_vec = np.zeros(self.n)  # process x (log stock)
        self.cur_time = 0

    def step(self, new_time):
        """Update x_vec in place when we move simulation by time dt."""

        dt = new_time - self.cur_time

        if callable(self.vol):
            vol = self.vol((self.cur_time, self.x_vec))
        else:
            vol = self.vol

        # generate the random numbers and advance the log stock process
        antithetic_normal(self.rng, self.n, sqrt(dt), self.dz_vec)
        self.dz_vec *= vol

        # add drift to x_vec: - vol * vol * dt/ 2.0
        np.multiply(vol, vol, out=self.tmp)
        self.tmp *= -0.5 * dt
        self.x_vec += self.tmp
        # add the random part to x_vec
        self.x_vec += self.dz_vec

        self.cur_time = new_time

    def get_value(self, unit):
        """Return the value of the modeled asset at the current time."""
        if unit == self.asset:
            return self.asset_fwd.forward(self.cur_time) * np.exp(self.x_vec)

    def get_df(self):
        """Return the discount factors at the current time."""
        return self.discounter.discount(self.cur_time)



"""
Testing ZCB for all models.
"""

import pytest
from pytest import approx

from finmc.calc.bond import zcb_price_mc
from finmc.utils.assets import Discounter
from tests.heston.dataset import data_heston_kruse
from tests.hullwhite.dataset import data_hwmc
from tests.localvol.dataset import (
    data_bsmc,
    data_lvmc,
    data_lvmc_fn,
    data_lvmc_grid,
)


@pytest.fixture(
    params=[
        data_bsmc,
        data_lvmc,
        data_lvmc_fn,
        data_lvmc_grid,
        data_heston_kruse,
        data_hwmc,
    ]
)
def data(request):
    return request.param()


@pytest.mark.parametrize("maturity", [0.1, 1.0, 3.0, 10.0])
def test_zcb(data, maturity):
    """Test the price of a zero coupon bond."""

    model_cls, dataset, _ = data
    if model_cls.__name__ == "HullWhiteMC":
        tol = 1e-3
    else:
        tol = 1e-5
        if maturity == 10.0:
            print("Skip.")
            return

    # Simulated Price of ZCB
    price = zcb_price_mc(
        maturity=maturity,
        asset_name="USD",
        model=model_cls(dataset),
    )

    # Get closed form price
    discounter = Discounter(dataset["ASSETS"][dataset["BASE"]])
    expected = discounter.discount(maturity)

    error = price - expected
    contract = f"ZCB {maturity:5.2f}"
    assert error == approx(0.0, abs=tol)

    print(f"{contract}: {price:11.6f} {expected:11.6f} {error * 1e4:7.2f}")



"""
Utility to calculate prices of bonds from a MC Simulation model.
"""

import numpy as np

from finmc.models.base import MCBase


def zcb_price_mc(
    maturity: float,
    asset_name: str,
    model: MCBase,
) -> float:
    """Calculate the price of a Zero Coupon Bond using MC Simulation.

    Args:
        maturity: The time to maturity of the bond in years.
        asset_name: The name of the asset.
        model: The model used to simulate the asset price.

    Returns:
        The price of the zero coupon bond.

    Examples:
        >>> price = zcb_price_mc(T, "USD", model)
    """

    model.reset()
    model.advance(maturity)
    df = model.get_df()

    if model.dataset.get("BASE") == asset_name:
        return df.mean()
    else:
        spots = model.get_value(asset_name)
        return (spots * df).mean()


if __name__ == "__main__":
    from finmc.models.hullwhite import HullWhiteMC

    dataset = {
        "MC": {
            "PATHS": 100_000,
            "TIMESTEP": 1 / 10,
        },
        "BASE": "USD",
        "ASSETS": {
            "USD": ("ZERO_RATES", np.array([[1.0, 0.04]])),
        },
        "HW": {
            "ASSET": "USD",
            "MEANREV": 0.1,
            "VOL": 0.03,
        },
    }

    model = HullWhiteMC(dataset)
    price = zcb_price_mc(
        maturity=1,
        asset_name="USD",
        model=model,
    )
    print(f"price = {price}")



# Description: This file is used to profile the code using line_profiler

from line_profiler import LineProfiler

from finmc.models.localvol import LVMC
from finmc.utils.mc import antithetic_normal
from tests.localvol.examples.option import run_model  # noqa: F401

if __name__ == "__main__":
    # Create a LineProfiler object, specifying the methods to be profiled by line
    lprofiler = LineProfiler(LVMC.step, antithetic_normal)
    # Wrap the entry function with the LineProfiler object, then run it.
    lp_wrapper = lprofiler(run_model)
    lp_wrapper()
    # Print the profiling results in milliseconds
    lprofiler.print_stats(output_unit=1e-03)



"""
This module contains utilities for Monte Carlo simulation.
"""

import numpy as np


def antithetic_normal(rng, n, scale, out):
    """Generate antithetic normal random numbers into a preallocated array.

    Args:
        rng: a random number generator.
        n: the number of samples to generate (must be even).
        scale: scaling factor to apply.
        out: the destination array of size n.

    Example:
        >>> rng = np.random.default_rng()
        >>> n = 10
        >>> out = np.empty(n, dtype=np.float64)
        >>> antithetic_normal(rng, n, 100.0, out)
        >>> print(out)
        [ -3.81721881   3.43045846 -25.60449457  90.48543061 -12.57430504
        3.81721881  -3.43045846  25.60449457 -90.48543061  12.57430504]

    """

    assert n % 2 == 0, "Number of paths must be even"
    h = n >> 1  # divide by 2
    rng.standard_normal(h, out=out[0:h])
    np.multiply(scale, out[0:h], out=out[0:h])
    np.negative(out[0:h], out=out[h:])


if __name__ == "__main__":
    rng = np.random.default_rng()
    n = 10
    out = np.empty(n, dtype=np.float64)
    antithetic_normal(rng, n, 100.0, out)
    print(out)
